import {
  getSettings,
  getSettingsByLine,
  getStrokeSettings,
} from "./settings.js";
import { transform } from "./transform.js";
import { circle, hline, vline, svg, text } from "../svg.js";

const drawMainLine = (settings) =>
  vline(
    settings.position.main.line,
    0,
    settings.size.height.main,
    getStrokeSettings({ settings, collateral: false })
  );

const drawAncestorCircle = ({ x, y, settings, collateral = false }) => {
  const circleSettings = getSettingsByLine(settings.circle, collateral);

  return circle(x, y, circleSettings.radius, {
    fill: "none",
    stroke: circleSettings.strokeColor,
    "stroke-width": circleSettings.strokeWidth,
  });
};

const drawAncestorInitial = ({
  x,
  y,
  initial,
  settings,
  collateral = false,
}) => {
  const textSettings = getSettingsByLine(settings.text, collateral);

  return text(x, y + 0.1, initial, {
    style: `alignment-baseline: middle; text-anchor: middle; font-size: ${textSettings.size}pt; fill: ${textSettings.fill};`,
  });
};

const drawAncestor = ({ x, y, initial, settings, collateral = false }) => [
  drawAncestorCircle({ x, y, settings, collateral }),
  drawAncestorInitial({ x, y, initial, settings, collateral }),
];

const drawGeneration = ({ gen, y, settings, collateral = false }) => {
  const position = collateral
    ? settings.position.collateral
    : settings.position.main;
  const [father_initial, mother_initial] = gen;

  return [
    ...(father_initial !== "-"
      ? drawAncestor({
          x: position.father,
          y,
          initial: father_initial,
          settings,
          collateral,
        })
      : []),
    ...(mother_initial !== "-"
      ? drawAncestor({
          x: position.mother,
          y,
          initial: mother_initial,
          settings,
          collateral,
        })
      : []),
  ];
};

const drawCollateralLines = ({ bar_y, settings }) => [
  hline(
    settings.position.main.line,
    settings.position.collateral.line -
      settings.line.collateral.strokeWidth / 2,
    bar_y,
    getStrokeSettings({ settings, collateral: false })
  ),
  vline(
    settings.position.collateral.line,
    bar_y,
    bar_y + settings.size.height.collateral,
    getStrokeSettings({ settings, collateral: true })
  ),
];

const render = (family) => {
  const { main, collateral } = transform(family);
  if (!main) return "";

  const settings = getSettings({ main, collateral });

  let y =
    settings.size.height.main -
    settings.circle.main.radius -
    settings.circle.main.strokeWidth;

  const output = [drawMainLine(settings)];

  for (const gen of main) {
    output.push(...drawGeneration({ gen, y, settings, collateral: false }));

    y -= settings.size.generation.height;
  }

  if (collateral.length > 0) {
    y += settings.size.generation.height;
    const bar_y = y + settings.size.generation.height / 2;
    y += settings.size.generation.height;

    output.push(...drawCollateralLines({ bar_y, settings }));

    for (const gen of collateral) {
      output.push(...drawGeneration({ gen, y, settings, collateral: true }));

      y += settings.size.generation.height;
    }
  }

  return svg({
    width: settings.size.width,
    height: settings.size.height.main,
    className: "ancestor-line",
    content: output.join("\n"),
  });
};

export default render;
