import * as fs from "node:fs";
import * as path from "node:path";
import { saveFile, getFiles, exportPDF } from "./files.js";
import { parseFile } from "./parse.js";
import { getDate, getTitle } from "./process.js";
import { renderSheet, renderTemplate } from "./render.js";
import renderAncestorLine from "./ancestorline/render.js";
import renderSparklines from "./sparklines/render.js";

import config from "../config.js";

const query = process.argv.length > 2 ? process.argv[2].toLowerCase() : null;

const loadFamilies = async (dataDirectoryPath, query) => {
  const families = [];
  const files = await getFiles(dataDirectoryPath);

  files.forEach((filename) => {
    if (query && !filename.includes(query)) return;

    const data = parseFile(filename);

    if (data.key) {
      data["filename"] = filename;
      families.push(data);
    } else {
      console.error(`${filename} is missing a key`);
    }
  });

  return families;
};

const families = await loadFamilies("data/", query);

const getFilenames = (family) => {
  const inputName = path.basename(family.filename).replace(".aml", "");

  const outputName = `${family.key} ${family.father.name ?? "Unknown"} and ${
    family.mother.name ?? "Unknown"
  }`;

  return {
    htmlFilename: `html/${inputName}.html`,
    pdfFilename: `pdf/${outputName}.pdf`,
  };
};

for (const family of families) {
  const { htmlFilename, pdfFilename } = getFilenames(family);

  const sheet = renderSheet(family);

  const title = getTitle(family);
  const date = getDate();
  const key = family.key;
  const ancestorLine = renderAncestorLine(family);
  const sparklines = renderSparklines(family);

  const outputData = renderTemplate({
    title,
    date,
    key,
    ancestorLine,
    sparklines,
    sheet,
  });

  saveFile(htmlFilename, outputData);

  if (config.pdfCommand && config.pathToHtml) {
    exportPDF({
      command: config.pdfCommand,
      path: `${config.pathToHtml}/${htmlFilename}`,
      filename: pdfFilename,
    });

    console.log(pdfFilename);
  } else {
    console.log(htmlFilename);
  }
}
