import { loadFile } from "./files.js";
import { ageAtEvent } from "./process.js";
import MarkdownIt from "markdown-it";

const renderName = (name, tag = "div") =>
  `<${tag} class="name">${name}</${tag}>`;

const renderNameSpan = (name) => renderName(name, "span");

const renderBulletList = (...parts) => parts.filter((part) => part).join(" • ");

const renderEvent = (event, abbr, birth) => {
  const age = ageAtEvent(birth, event);
  const ageString = age !== null ? ` at age ${age}` : "";

  return event ? `${abbr}. ${event}${ageString}` : "";
};

const renderLink = (link) => (link ? `<span class="link">${link}</span>` : "");

const renderFamilySearchId = (person) =>
  person.id ? `<span class="fsid">${person.id}</span>` : "";

const renderLifeLine = (person) => {
  return `<div class="metadata life">${renderBulletList(
    renderEvent(person.birth, "b"),
    renderEvent(person.death, "d", person.birth),
    renderFamilySearchId(person)
  )}</div>`;
};

const renderParentParents = (parents) => {
  if (!parents) return "";

  const parentsLine = `Parents: ${renderNameSpan(
    parents.father ?? "Unknown"
  )} &amp; ${renderNameSpan(parents.mother ?? "Unknown")}`;

  return `<div class="metadata parents">${renderBulletList(
    parentsLine,
    renderLink(parents.link)
  )}</div>`;
};

const renderParent = (parent) => {
  if (!parent.name && !parent.life) return "";

  return `<div class="person parent">
  ${renderName(parent.name)}
  ${renderLifeLine(parent)}
  ${renderParentParents(parent.parents)}
</div>`;
};

const renderParents = (family) => {
  if (!family.father && !family.mother) return "";

  return `<div class="parents">${renderParent(family.father)}${renderParent(
    family.mother
  )}</div>`;
};

const renderMarriage = (family) => {
  if (!family.marriage) return "";

  const marriageEvent = [family.marriage.date, family.marriage.place]
    .filter((part) => part)
    .join(" in ");

  const husbandAge = ageAtEvent(family.father.birth, family.marriage.date);
  const wifeAge = ageAtEvent(family.mother.birth, family.marriage.date);

  const ages =
    husbandAge || wifeAge
      ? `ages ${husbandAge ?? "?"} and ${wifeAge ?? "?"}`
      : "";

  const divorce = family.marriage.divorce
    ? `divorced ${family.marriage.divorce}`
    : "";

  return `<div class="marriage"><div class="metadata life">${renderBulletList(
    renderEvent(marriageEvent, "m"),
    ages,
    divorce,
    family.marriage.note
  )}</div></div>`;
};

const renderChildSpouse = (spouse) => {
  const spouseLine = `Spouse: ${renderNameSpan(spouse.name)}`;
  const marriageDate = spouse.date ? `m. ${spouse.date}` : "";
  const divorceDate = spouse.divorce ? `divorced ${spouse.divorce}` : "";

  const numChildren = spouse.children?.length ?? 0;
  const children = numChildren
    ? `${numChildren} child${numChildren !== 1 ? "ren" : ""}`
    : "";

  return `<div class="metadata spouse">
        ${renderBulletList(
          spouseLine,
          marriageDate,
          divorceDate,
          children,
          renderLink(spouse.link)
        )}
          </div>`;
};

const renderChild = (child) => {
  if (!child) return "";

  return `<div class="person child">
${renderName(child.name)}
${renderLifeLine(child)}
${(child.spouses ?? []).map(renderChildSpouse).join("\n")}
</div>`;
};

const renderChildren = (family) => {
  if (!family.children || family.children.length === 0) return "";

  return `<section>
      <h2>Children</h2>

      ${family.children.map(renderChild).join("\n")}
</section>`;
};

const renderNotes = (notes) => {
  if (!notes) return "";

  const md = new MarkdownIt({ typographer: true });

  return `<section class="notes">
  <h2>Notes</h2>
  ${md.render(notes)}
</section>`;
};

export const renderSheet = (family) => {
  return `${renderParents(family)}
${renderMarriage(family)}
${renderChildren(family)}
${renderNotes(family.notes)}`;
};

export const renderTemplate = (variables) => {
  let template = loadFile("template/index.html");

  Object.keys(variables).forEach((key) => {
    template = template.replaceAll(`{{ ${key} }}`, variables[key]);
  });

  return template;
};
