import { parseEventString } from "./parse.js";

const getLabel = (person) => person.name.replace(/[^A-Z]/g, "") ?? "?";

const getSegmentForEvent = (event) => {
  if (!event) return null;

  if (event.after || event.before) {
    return {
      type: "dotted",
      start: event.after.x,
      end: event.before.x,
    };
  }
};

const eventIsDefinite = (event) => event.year && !event.birth && !event.after;

const getSegmentsForPerson = (person) => {
  const birthSegment = getSegmentForEvent(person.birth);
  const deathSegment = getSegmentForEvent(person.death);

  const mainSegment = {
    type: "solid",
    start: birthSegment ? birthSegment.end : person.birth.x,
    end: deathSegment ? deathSegment.start : person.death.x,
  };

  return {
    segments: [birthSegment, mainSegment, deathSegment].filter((s) => s),
    mainSegment,
  };
};

const getTrack = (person, type) => {
  const { segments, mainSegment } = getSegmentsForPerson(person);

  if (!isFinite(mainSegment.start) || !isFinite(mainSegment.end)) {
    return null;
  }

  return {
    ...person,
    type,
    label: getLabel(person),
    segments,
    markers: getMarkersForPerson(person, mainSegment),
  };
};

// Pull this marriage/children into the parent object
const augmentParent = (parent, marriage, children) => {
  return {
    ...parent,
    spouses: [
      ...(marriage || children
        ? [
            {
              ...(marriage?.date ? { date: marriage.date } : {}),
              ...(marriage?.divorce ? { divorce: marriage.divorce } : {}),
              children: children.map((child) => ({
                birth: child.birth ?? "unknown",
                gender: child.gender ?? "unknown",
              })),
            },
          ]
        : []),
      ...(parent.spouses ?? []),
    ],
  };
};

const getTracks = (family) => {
  return [
    getTrack(family.father, "parent"),
    ...family.children.map((child) => getTrack(child, "child")),
    getTrack(family.mother, "parent"),
  ].filter((track) => track);
};

// Markers

const getMarker = (x, type) => ({ x, type });

const getMarkerFromChildGender = (child) =>
  ({ m: "son", f: "daughter" }[child.gender] ?? "child");

const getMarkersForSpouseMarriages = (person) =>
  (person.spouses || [])
    .map((spouse) => (spouse.x ? getMarker(spouse.x, "marriage") : null))
    .flat();

const getMarkersForSpouseDivorces = (person) =>
  (person.spouses || [])
    .map((spouse) =>
      spouse.divorce?.x ? getMarker(spouse.divorce.x, "divorce") : null
    )
    .flat();

const childrenWithX = (person) =>
  (person.children || []).filter((child) => child.birth?.x);

const getMarkersForChildren = (person) =>
  (person.spouses || [])
    .map((spouse) => {
      return childrenWithX(spouse).map((child) =>
        getMarker(child.birth.x, getMarkerFromChildGender(child))
      );
    })
    .flat();

const getMarkersForPerson = (person) => {
  // Decade marks
  let tickMarks = [];
  if (eventIsDefinite(person.birth)) {
    for (let dx = person.birth.x; dx < person.death.x; dx += person.decadeX) {
      tickMarks.push(getMarker(dx, "tick"));
    }
  }

  const markers = [
    eventIsDefinite(person.birth) ? getMarker(person.birth.x, "birth") : null,
    eventIsDefinite(person.death) ? getMarker(person.death.x, "death") : null,
    ...getMarkersForSpouseMarriages(person),
    ...getMarkersForSpouseDivorces(person),
    ...getMarkersForChildren(person),
  ].filter((m) => m);

  return [...tickMarks, ...markers];
};

// Events

const transformEventDate = (object) => ({
  ...object,
  ...(object?.date ? parseEventString(object.date) : {}),
});

const transformEventsForSpouse = (spouse) => ({
  ...spouse,
  ...(spouse?.date ? parseEventString(spouse.date) : {}),
  ...(spouse?.divorce ? { divorce: parseEventString(spouse.divorce) } : {}),
  ...(spouse?.children
    ? { children: spouse.children.map(transformEventsForPerson) }
    : {}),
});

const transformEventsForPerson = (person) => {
  return {
    ...person,
    ...(person?.birth ? { birth: parseEventString(person.birth) } : {}),
    ...(person?.death ? { death: parseEventString(person.death) } : {}),
    ...((person?.spouses ?? []).length
      ? { spouses: person.spouses.map(transformEventsForSpouse) }
      : {}),
  };
};

const transformEventsForFamily = (family) => {
  const father = augmentParent(family.father, family.marriage, family.children);
  const mother = augmentParent(family.mother, family.marriage, family.children);

  return {
    ...family,
    father: transformEventsForPerson(father),
    mother: transformEventsForPerson(mother),
    marriage: transformEventDate(family.marriage),
    children: family.children.map(transformEventsForPerson),
  };
};

// Coordinates

const getYearsForEvent = (event, add = 0) => [
  event?.year + add,
  event?.after?.year + add,
  event?.before?.year + add,
];

const getChildrenYears = (children) =>
  (children || []).map((child) => getFirstYearForPerson(child)).flat();

const getSpouseYears = (spouse) => [
  spouse.year,
  ...getChildrenYears(spouse.children),
];

const getFirstYearForFamily = (family) => {
  // Note: we've already augmented by this point, so the father and mother have the marriage and children already
  const eventYears = [
    getFirstYearForPerson(family.father),
    getFirstYearForPerson(family.mother),
    ...(family.father?.spouses || []).map(getSpouseYears),
    ...(family.mother?.spouses || []).map(getSpouseYears),
  ]
    .flat()
    .filter((y) => y);

  return Math.min(...eventYears);
};

const getFirstYearForPerson = (person) => {
  const eventYears = [
    ...getYearsForEvent(person.birth),
    ...getYearsForEvent(person.death, -5),
    ...(person.spouses || [])
      .map((spouse) => getYearsForEvent(spouse, -20))
      .flat(),
    ...(person.spouses || []).map((spouse) =>
      (spouse.children || [])
        .map((child) => getYearsForEvent(child.birth, -20))
        .flat()
    ),
  ]
    .flat()
    .filter((y) => y);

  return Math.min(...eventYears);
};

const getLastYearForPerson = (person) => {
  if (person.alive) {
    return new Date().getFullYear();
  }

  const eventYears = [
    ...getYearsForEvent(person.birth, 5),
    ...getYearsForEvent(person.death),
    ...(person.spouses || []).map((spouse) => getYearsForEvent(spouse)),
    ...(person.spouses || []).map((spouse) =>
      (spouse.children || [])
        .map((child) => getYearsForEvent(child.birth))
        .flat()
    ),
  ]
    .flat()
    .filter((y) => y);

  return Math.max(...eventYears);
};

const daysPerMonth = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31];

const getXFromEvent = (event, firstYear, xFactor) => {
  // If there's no year, just return the first year since there's nothing we can really do
  if (event.year === undefined) {
    return 0;
  }

  let x = event.year - firstYear;

  if (event.month !== undefined) {
    x += event.month / 12;

    // Only do day if we have the month
    if (event.day !== undefined) {
      // Divide by 12 so that it's within a month
      x += event.day / daysPerMonth[event.month - 1] / 12;
    }
  }

  return x * (xFactor ?? 1);
};

const transformCoordinatesForDate = (object, firstYear, xFactor) => ({
  ...object,
  x: getXFromEvent(object, firstYear, xFactor),
  ...(object.before
    ? {
        before: {
          ...object.before,
          x: getXFromEvent(object.before, firstYear, xFactor),
        },
      }
    : {}),
  ...(object.after
    ? {
        after: {
          ...object.after,
          x: getXFromEvent(object.after, firstYear, xFactor),
        },
      }
    : {}),
});

const transformCoordinatesForSpouse = (spouse, firstYear, xFactor) => ({
  ...spouse,
  ...transformCoordinatesForDate(spouse, firstYear, xFactor),
  ...(spouse.divorce
    ? {
        divorce: transformCoordinatesForDate(
          spouse.divorce,
          firstYear,
          xFactor
        ),
      }
    : {}),
  ...(spouse.children
    ? {
        children: spouse.children.map((child) =>
          transformCoordinatesForPerson(child, firstYear, xFactor)
        ),
      }
    : {}),
});

const transformBirthCoordinates = (person, firstYear, xFactor) => ({
  birth: person.birth
    ? transformCoordinatesForDate(person.birth, firstYear, xFactor)
    : {
        x: getXFromEvent(
          // Go with January 1 so we cover the year
          { year: getFirstYearForPerson(person), month: 1, day: 1 },
          firstYear,
          xFactor
        ),
      },
});

const transformDeathCoordinates = (person, firstYear, xFactor) => ({
  death: person.death
    ? transformCoordinatesForDate(person.death, firstYear, xFactor)
    : {
        x: getXFromEvent(
          // Go with December 31 so we cover the year
          {
            year: getLastYearForPerson(person),
            month: 11, // December
            day: 31,
          },
          firstYear,
          xFactor
        ),
      },
});

const transformSpouseCoordinates = (person, firstYear, xFactor) =>
  (person.spouses ?? []).length
    ? {
        spouses: person.spouses.map((spouse) =>
          transformCoordinatesForSpouse(spouse, firstYear, xFactor)
        ),
      }
    : {};

const transformCoordinatesForPerson = (person, firstYear, xFactor) => {
  return {
    ...person,
    ...transformBirthCoordinates(person, firstYear, xFactor),
    ...transformDeathCoordinates(person, firstYear, xFactor),
    ...transformSpouseCoordinates(person, firstYear, xFactor),
    decadeX: 10 * xFactor, // Used for tick marks
  };
};

const transformCoordinatesForFamily = (family, settings) => {
  const xFactor = settings.size.xFactor ?? 1;
  const firstYear = getFirstYearForFamily(family);

  return {
    ...family,
    father: transformCoordinatesForPerson(family.father, firstYear, xFactor),
    mother: transformCoordinatesForPerson(family.mother, firstYear, xFactor),
    marriage: transformCoordinatesForDate(family.marriage, firstYear, xFactor),
    children: family.children.map((child) =>
      transformCoordinatesForPerson(child, firstYear, xFactor)
    ),
  };
};

// Bands

const getBands = (family, settings) => {
  const numTracks = [family.father, family.mother, ...family.children].filter(
    (p) => p
  ).length;

  if (family.marriage.x) {
    return [
      {
        x: family.marriage.x + (settings.bands["stroke-width"] / 2 ?? 0),
        start: 0,
        end: numTracks - 1,
        type: "solid",
      },
    ];
  }

  return [];
};

// Put it all together

export const transform = (family, settings) => {
  // Parse out the dates and places
  const familyWithEvents = transformEventsForFamily(family);

  // Now that we have all the events ready, transform them into scaled X coordinates
  const familyWithCoordinates = transformCoordinatesForFamily(
    familyWithEvents,
    settings
  );

  const response = {
    tracks: getTracks(familyWithCoordinates),
    bands: getBands(familyWithCoordinates, settings),
  };

  return response;
};
